from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SchedulerConfig(AppConfig):
    default_auto_field = 'django.db.models.AutoField'
    name = 'scheduler'
    verbose_name = _('Tasks Scheduler')

    def ready(self):
        pass



from datetime import timedelta

from django.utils import timezone

from scheduler.models import ScheduledJob
from scheduler.tests.testtools import SchedulerBaseCase, job_factory
from scheduler.tools import get_scheduled_job


class TestInternals(SchedulerBaseCase):
    def test_get_scheduled_job(self):
        job = job_factory(ScheduledJob, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertEqual(job, get_scheduled_job(job.JOB_TYPE, job.id))
        with self.assertRaises(ValueError):
            get_scheduled_job(job.JOB_TYPE, job.id + 1)
        with self.assertRaises(ValueError):
            get_scheduled_job('UNKNOWN_JOBTYPE', job.id)



import os

from django.conf import settings

from scheduler.settings import conf_settings

settings.SCHEDULER_QUEUES = {
    'default': {'HOST': 'localhost', 'PORT': 6379, 'DB': 0, 'DEFAULT_TIMEOUT': 500},
    'test': {
        'HOST': 'localhost',
        'PORT': 1,
        'DB': 1,
    },
    'sentinel': {
        'SENTINELS': [('localhost', 26736), ('localhost', 26737)],
        'MASTER_NAME': 'testmaster',
        'DB': 1,
        'USERNAME': 'redis-user',
        'PASSWORD': 'secret',
        'SOCKET_TIMEOUT': 10,
        'SENTINEL_KWARGS': {},
    },
    'test1': {
        'HOST': 'localhost',
        'PORT': 1,
        'DB': 1,
        'DEFAULT_TIMEOUT': 400,
    },
    'test2': {
        'HOST': 'localhost',
        'PORT': 1,
        'DB': 1,
    },
    'test3': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 1,
    },
    'async': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 1,
        'ASYNC': False,
    },
    'url': {
        'URL': 'redis://username:password@host:1234/',
        'DB': 4,
    },
    'url_with_db': {
        'URL': 'redis://username:password@host:1234/5',
    },
    'url_default_db': {
        'URL': 'redis://username:password@host:1234',
    },
    'django_tasks_scheduler_test': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
    },
    'scheduler_scheduler_active_test': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'ASYNC': False,
    },
    'scheduler_scheduler_inactive_test': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'ASYNC': False,
    },
    'worker_scheduler_active_test': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'ASYNC': False,
    },
    'worker_scheduler_inactive_test': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'ASYNC': False,
    },
    'django_tasks_scheduler_test2': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
    },
    'test_scheduler': {
        'HOST': 'localhost',
        'PORT': 6379,
        'DB': 0,
        'DEFAULT_TIMEOUT': 400,
    },
}
settings.SCHEDULER_CONFIG = dict(
    FAKEREDIS=(os.getenv('FAKEREDIS', 'False') == 'True'),
)
conf_settings()



from datetime import timedelta

from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.messages import get_messages
from django.test import Client, TestCase
from django.utils import timezone

from scheduler import settings
from scheduler.models import CronJob, JobKwarg, RepeatableJob, ScheduledJob, BaseJob
from scheduler.queues import get_queue


def assert_message_in_response(response, message):
    messages = [m.message for m in get_messages(response.wsgi_request)]
    assert message in messages, f'expected "{message}" in {messages}'


def sequence_gen():
    n = 1
    while True:
        yield n
        n += 1


seq = sequence_gen()


def job_factory(cls, instance_only=False, **kwargs):
    values = dict(
        name='Scheduled Job %d' % next(seq),
        job_id=None,
        queue=list(settings.QUEUES.keys())[0],
        callable='scheduler.tests.jobs.test_job',
        enabled=True,
        timeout=None)
    if cls == ScheduledJob:
        values.update(dict(
            result_ttl=None,
            scheduled_time=timezone.now() + timedelta(days=1), ))
    elif cls == RepeatableJob:
        values.update(dict(
            result_ttl=None,
            interval=1,
            interval_unit='hours',
            repeat=None,
            scheduled_time=timezone.now() + timedelta(days=1), ))
    elif cls == CronJob:
        values.update(dict(cron_string="0 0 * * *", repeat=None, ))
    values.update(kwargs)
    if instance_only:
        instance = cls(**values)
    else:
        instance = cls.objects.create(**values)
    return instance


def jobarg_factory(cls, **kwargs):
    content_object = kwargs.pop('content_object', None)
    if content_object is None:
        content_object = job_factory(ScheduledJob)
    values = dict(
        arg_type='str',
        val='',
        object_id=content_object.id,
        content_type=ContentType.objects.get_for_model(content_object),
        content_object=content_object,
    )
    if cls == JobKwarg:
        values['key'] = 'key%d' % next(seq),
    values.update(kwargs)
    instance = cls.objects.create(**values)
    return instance


def _get_job_from_scheduled_registry(django_job: BaseJob):
    jobs_to_schedule = django_job.rqueue.scheduled_job_registry.get_job_ids()
    entry = next(i for i in jobs_to_schedule if i == django_job.job_id)
    return django_job.rqueue.fetch_job(entry)


def _get_executions(django_job: BaseJob):
    job_ids = django_job.rqueue.get_all_job_ids()
    return list(filter(
        lambda j: j.is_execution_of(django_job),
        map(lambda jid: django_job.rqueue.fetch_job(jid), job_ids)))


class SchedulerBaseCase(TestCase):
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        try:
            User.objects.create_superuser('admin', 'admin@a.com', 'admin')
        except Exception:
            pass
        cls.client = Client()

    def setUp(self) -> None:
        super(SchedulerBaseCase, self).setUp()
        queue = get_queue('default')
        queue.empty()

    def tearDown(self) -> None:
        super(SchedulerBaseCase, self).setUp()
        queue = get_queue('default')
        queue.empty()

    @classmethod
    def setUpClass(cls):
        super(SchedulerBaseCase, cls).setUpClass()
        queue = get_queue('default')
        queue.connection.flushall()



from typing import Dict

from django import template
from django.utils.safestring import mark_safe

from scheduler.rq_classes import JobExecution, DjangoQueue
from scheduler.tools import get_scheduled_job

register = template.Library()


@register.filter
def show_func_name(rq_job: JobExecution) -> str:
    try:
        res = rq_job.func_name
        if res == 'scheduler.tools.run_job':
            job = get_scheduled_job(*rq_job.args)
            res = job.function_string()
        return mark_safe(res)
    except Exception as e:
        return repr(e)


@register.filter
def get_item(dictionary: Dict, key):
    return dictionary.get(key)


@register.filter
def scheduled_job(job: JobExecution):
    django_scheduled_job = get_scheduled_job(*job.args)
    return django_scheduled_job.get_absolute_url()


@register.filter
def worker_scheduler_pid(worker):
    return worker.scheduler_pid()


@register.filter
def job_result(job: JobExecution):
    result = job.latest_result()
    return result.type.name.capitalize() if result else None


@register.filter
def job_status(job: JobExecution):
    result = job.get_status()
    return result.capitalize()


@register.filter
def job_runtime(job: JobExecution):
    ended_at = job.ended_at
    if ended_at:
        runtime = job.ended_at - job.started_at
        return f'{int(runtime.microseconds / 1000)}ms'
    elif job.started_at:
        return "Still running"
    else:
        return "-"


@register.filter
def job_scheduled_time(job: JobExecution, queue: DjangoQueue):
    try:
        return queue.scheduled_job_registry.get_scheduled_time(job.id)
    except Exception:
        return None



from django.db import models


class Queue(models.Model):
    """Placeholder model with no database table, but with django admin page
    and contenttype permission"""

    class Meta:
        managed = False  # not in Django's database
        default_permissions = ()
        permissions = [['view', 'Access admin page']]
        verbose_name_plural = " Queues"



import click
from django.core.management.base import BaseCommand

from scheduler.queues import get_queue
from scheduler.rq_classes import JobExecution


class Command(BaseCommand):
    help = 'Delete failed jobs from Django queue.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--queue', '-q', dest='queue', default='default',
            help='Specify the queue [default]')
        parser.add_argument('-f', '--func', help='optional job function name, e.g. "app.tasks.func"')
        parser.add_argument('--dry-run', action='store_true', help='Do not actually delete failed jobs')

    def handle(self, *args, **options):
        queue = get_queue(options.get('queue', 'default'))
        job_ids = queue.failed_job_registry.get_job_ids()
        jobs = [JobExecution.fetch(job_id, connection=queue.connection) for job_id in job_ids]
        func_name = options.get('func', None)
        if func_name is not None:
            jobs = [job for job in jobs if job.func_name == func_name]
        dry_run = options.get('dry_run', False)
        click.echo(f'Found {len(jobs)} failed jobs')
        for job in jobs:
            click.echo(f'Deleting {job.id}')
            if not dry_run:
                job.delete()
        click.echo(f'Deleted {len(jobs)} failed jobs')



from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from scheduler.models import JobArg, JobKwarg
from .jobs import arg_callable
from .testtools import jobarg_factory


class TestAllJobArg(TestCase):
    JobArgClass = JobArg

    def test_bad_arg_type(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='bad_arg_type', val='something')
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_one_value_invalid_str_int(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='int', val='not blank', )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_callable_invalid(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='callable', val='bad_callable', )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_datetime_invalid(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='datetime', val='bad datetime', )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_bool_invalid(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='bool', val='bad bool', )
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_clean_int_invalid(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='int', val='str')
        with self.assertRaises(ValidationError):
            arg.clean()

    def test_str_clean(self):
        arg = jobarg_factory(self.JobArgClass, val='something')
        self.assertIsNone(arg.clean())


class TestJobArg(TestCase):
    JobArgClass = JobArg

    def test_str(self):
        arg = jobarg_factory(self.JobArgClass)
        self.assertEqual(
            f'JobArg[arg_type={arg.arg_type},value={arg.value()}]', str(arg))

    def test_value(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='str', val='something')
        self.assertEqual(arg.value(), 'something')

    def test__str__str_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='str', val='something')
        self.assertEqual('something', str(arg.value()))

    def test__str__int_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='int', val='1')
        self.assertEqual('1', str(arg.value()))

    def test__str__datetime_val(self):
        _time = timezone.now()
        arg = jobarg_factory(self.JobArgClass, arg_type='datetime', val=str(_time))
        self.assertEqual(str(_time), str(arg.value()))

    def test__str__bool_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='bool', val='True')
        self.assertEqual('True', str(arg.value()))

    def test__repr__str_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='str', val='something')
        self.assertEqual("'something'", repr(arg.value()))

    def test__repr__int_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='int', val='1')
        self.assertEqual('1', repr(arg.value()))

    def test__repr__datetime_val(self):
        _time = timezone.now()
        arg = jobarg_factory(self.JobArgClass, arg_type='datetime', val=str(_time))
        self.assertEqual(repr(_time), repr(arg.value()))

    def test__repr__bool_val(self):
        arg = jobarg_factory(self.JobArgClass, arg_type='bool', val='False')
        self.assertEqual('False', repr(arg.value()))

    def test_callable_arg_type__clean(self):
        method = arg_callable
        arg = jobarg_factory(
            self.JobArgClass, arg_type='callable',
            val=f'{method.__module__}.{method.__name__}', )
        self.assertIsNone(arg.clean())
        self.assertEqual(1, arg.value())
        self.assertEqual(2, arg.value())


class TestJobKwarg(TestAllJobArg):
    JobArgClass = JobKwarg

    def test_str(self):
        arg = jobarg_factory(self.JobArgClass)
        self.assertEqual(
            f'JobKwarg[key={arg.key},arg_type={arg.arg_type},value={arg.val}]', str(arg))

    def test_value(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='str', val='value')
        self.assertEqual(kwarg.value(), ('key', 'value'))

    def test__str__str_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='str', val='something')
        self.assertEqual('JobKwarg[key=key,arg_type=str,value=something]', str(kwarg))

    def test__str__int_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='int', val=1)
        self.assertEqual('JobKwarg[key=key,arg_type=int,value=1]', str(kwarg))

    def test__str__datetime_val(self):
        _time = timezone.now()
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='datetime', val=str(_time))
        self.assertEqual(f'JobKwarg[key=key,arg_type=datetime,value={_time}]', str(kwarg))

    def test__str__bool_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='bool', val='True')
        self.assertEqual('JobKwarg[key=key,arg_type=bool,value=True]', str(kwarg))

    def test__repr__str_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='str', val='something')
        self.assertEqual("('key', 'something')", repr(kwarg.value()))

    def test__repr__int_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='int', val='1')
        self.assertEqual("('key', 1)", repr(kwarg.value()))

    def test__repr__datetime_val(self):
        _time = timezone.now()
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='datetime', val=str(_time))
        self.assertEqual("('key', {})".format(repr(_time)), repr(kwarg.value()))

    def test__repr__bool_val(self):
        kwarg = jobarg_factory(self.JobArgClass, key='key', arg_type='bool', val='True')
        self.assertEqual("('key', True)", repr(kwarg.value()))



# Generated by Django 4.0.1 on 2022-01-06 20:34

from django.db import migrations, models


class Migration(migrations.Migration):
    # replaces = [('scheduler', '0001_initial'), ('scheduler', '0002_add_timeout'),
    #             ('scheduler', '0003_remove_queue_choices'), ('scheduler', '0004_add_cron_jobs'),
    #             ('scheduler', '0005_added_result_ttl')]

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CronJob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='name')),
                ('callable', models.CharField(max_length=2048, verbose_name='callable')),
                ('enabled', models.BooleanField(default=True, verbose_name='enabled')),
                ('queue', models.CharField(max_length=16, verbose_name='queue')),
                ('job_id',
                 models.CharField(blank=True, editable=False, max_length=128, null=True, verbose_name='job id')),
                ('timeout', models.IntegerField(blank=True,
                                                help_text="Timeout specifies the maximum runtime, in seconds, for the job before it'll be considered 'lost'. Blank uses the default timeout.",
                                                null=True, verbose_name='timeout')),
                ('repeat', models.PositiveIntegerField(blank=True, null=True, verbose_name='repeat')),
                ('cron_string',
                 models.CharField(help_text='Define the schedule in a crontab like syntax.', max_length=64,
                                  verbose_name='cron string')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'Cron Job',
                'verbose_name_plural': 'Cron Jobs',
            },
        ),
        migrations.CreateModel(
            name='RepeatableJob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='name')),
                ('callable', models.CharField(max_length=2048, verbose_name='callable')),
                ('enabled', models.BooleanField(default=True, verbose_name='enabled')),
                ('queue', models.CharField(max_length=16, verbose_name='queue')),
                ('job_id',
                 models.CharField(blank=True, editable=False, max_length=128, null=True, verbose_name='job id')),
                ('scheduled_time', models.DateTimeField(verbose_name='scheduled time')),
                ('interval', models.PositiveIntegerField(verbose_name='interval')),
                ('interval_unit', models.CharField(
                    choices=[('minutes', 'minutes'), ('hours', 'hours'), ('days', 'days'), ('weeks', 'weeks')],
                    default='hours', max_length=12, verbose_name='interval unit')),
                ('repeat', models.PositiveIntegerField(blank=True, null=True, verbose_name='repeat')),
                ('timeout', models.IntegerField(blank=True,
                                                help_text="Timeout specifies the maximum runtime, in seconds, for the job before it'll be considered 'lost'. Blank uses the default timeout.",
                                                null=True, verbose_name='timeout')),
                ('result_ttl', models.IntegerField(blank=True,
                                                   help_text='The TTL value (in seconds) of the job result. -1: Result never expires, you should delete jobs manually. 0: Result gets deleted immediately. >0: Result expires after n seconds.',
                                                   null=True, verbose_name='result ttl')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'Repeatable Job',
                'verbose_name_plural': 'Repeatable Jobs',
            },
        ),
        migrations.CreateModel(
            name='ScheduledJob',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', models.DateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(max_length=128, unique=True, verbose_name='name')),
                ('callable', models.CharField(max_length=2048, verbose_name='callable')),
                ('enabled', models.BooleanField(default=True, verbose_name='enabled')),
                ('queue', models.CharField(max_length=16, verbose_name='queue')),
                ('job_id',
                 models.CharField(blank=True, editable=False, max_length=128, null=True, verbose_name='job id')),
                ('scheduled_time', models.DateTimeField(verbose_name='scheduled time')),
                ('timeout', models.IntegerField(blank=True,
                                                help_text="Timeout specifies the maximum runtime, in seconds, for the job before it'll be considered 'lost'. Blank uses the default timeout.",
                                                null=True, verbose_name='timeout')),
                ('result_ttl', models.IntegerField(blank=True,
                                                   help_text='The TTL value (in seconds) of the job result. -1: Result never expires, you should delete jobs manually. 0: Result gets deleted immediately. >0: Result expires after n seconds.',
                                                   null=True, verbose_name='result ttl')),
            ],
            options={
                'ordering': ('name',),
                'verbose_name': 'Scheduled Job',
                'verbose_name_plural': 'Scheduled Jobs',
            },
        ),
    ]



import os
import uuid

from scheduler.tests.testtools import SchedulerBaseCase
from scheduler.tools import create_worker
from . import test_settings  # noqa
from .. import settings


class TestWorker(SchedulerBaseCase):
    def test_create_worker__two_workers_same_queue(self):
        worker1 = create_worker('default', 'django_tasks_scheduler_test')
        worker1.register_birth()
        worker2 = create_worker('default')
        worker2.register_birth()
        hostname = os.uname()[1]
        self.assertEqual(f'{hostname}-worker.1', worker1.name)
        self.assertEqual(f'{hostname}-worker.2', worker2.name)

    def test_create_worker__worker_with_queues_different_connection(self):
        with self.assertRaises(ValueError):
            create_worker('default', 'test1')

    def test_create_worker__with_name(self):
        name = uuid.uuid4().hex
        worker1 = create_worker('default', name=name)
        self.assertEqual(name, worker1.name)

    def test_create_worker__with_name_containing_slash(self):
        name = uuid.uuid4().hex[-4:] + '/' + uuid.uuid4().hex[-4:]
        worker1 = create_worker('default', name=name)
        self.assertEqual(name.replace('/', '.'), worker1.name)

    def test_create_worker__scheduler_interval(self):
        prev = settings.SCHEDULER_CONFIG['SCHEDULER_INTERVAL']
        settings.SCHEDULER_CONFIG['SCHEDULER_INTERVAL'] = 1
        worker = create_worker('default')
        worker.work(burst=True)
        self.assertEqual(worker.scheduler.interval, 1)
        settings.SCHEDULER_CONFIG['SCHEDULER_INTERVAL'] = prev



# Generated by Django 4.1.7 on 2023-03-12 19:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('scheduler', '0008_rename_str_val_jobarg_val_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='jobarg',
            name='arg_type',
            field=models.CharField(
                choices=[('str', 'string'), ('int', 'int'), ('bool', 'boolean'), ('datetime', 'datetime'),
                         ('callable', 'callable')], default='str', max_length=12, verbose_name='Argument Type'),
        ),
        migrations.AlterField(
            model_name='jobarg',
            name='val',
            field=models.CharField(blank=True, max_length=255, verbose_name='Argument Value'),
        ),
        migrations.AlterField(
            model_name='jobkwarg',
            name='arg_type',
            field=models.CharField(
                choices=[('str', 'string'), ('int', 'int'), ('bool', 'boolean'), ('datetime', 'datetime'),
                         ('callable', 'callable')], default='str', max_length=12, verbose_name='Argument Type'),
        ),
        migrations.AlterField(
            model_name='jobkwarg',
            name='val',
            field=models.CharField(blank=True, max_length=255, verbose_name='Argument Value'),
        ),
    ]



import zoneinfo
from datetime import datetime, timedelta

from django.contrib.messages import get_messages
from django.core.exceptions import ValidationError
from django.test import override_settings
from django.urls import reverse
from django.utils import timezone

from scheduler import settings
from scheduler.models import BaseJob, CronJob, JobArg, JobKwarg, RepeatableJob, ScheduledJob
from scheduler.tools import run_job, create_worker
from . import jobs
from .testtools import (
    job_factory, jobarg_factory, _get_job_from_scheduled_registry,
    SchedulerBaseCase, _get_executions)
from ..queues import get_queue


def assert_response_has_msg(response, message):
    messages = [m.message for m in get_messages(response.wsgi_request)]
    assert message in messages, f'expected "{message}" in {messages}'


def has_execution_with_status(django_job, status) -> bool:
    job_list = _get_executions(django_job)
    job_list = [(j.id, j.get_status()) for j in job_list]
    for job in job_list:
        if job[1] == status:
            return True
    print(f'{job_list} does not contain a job with status {status}')
    return False


class BaseTestCases:
    class TestBaseJob(SchedulerBaseCase):
        JobModelClass = BaseJob

        def test_callable_func(self):
            job = job_factory(self.JobModelClass)
            job.callable = 'scheduler.tests.jobs.test_job'
            func = job.callable_func()
            self.assertEqual(jobs.test_job, func)

        def test_callable_func_not_callable(self):
            job = job_factory(self.JobModelClass)
            job.callable = 'scheduler.tests.jobs.test_non_callable'
            with self.assertRaises(TypeError):
                job.callable_func()

        def test_clean_callable(self):
            job = job_factory(self.JobModelClass)
            job.callable = 'scheduler.tests.jobs.test_job'
            self.assertIsNone(job.clean_callable())

        def test_clean_callable_invalid(self):
            job = job_factory(self.JobModelClass)
            job.callable = 'scheduler.tests.jobs.test_non_callable'
            with self.assertRaises(ValidationError):
                job.clean_callable()

        def test_clean_queue(self):
            for queue in settings.QUEUES.keys():
                job = job_factory(self.JobModelClass)
                job.queue = queue
                self.assertIsNone(job.clean_queue())

        def test_clean_queue_invalid(self):
            job = job_factory(self.JobModelClass)
            job.queue = 'xxxxxx'
            job.callable = 'scheduler.tests.jobs.test_job'
            with self.assertRaises(ValidationError):
                job.clean()

        # next 2 check the above are included in job.clean() function
        def test_clean_base(self):
            job = job_factory(self.JobModelClass)
            job.queue = list(settings.QUEUES)[0]
            job.callable = 'scheduler.tests.jobs.test_job'
            self.assertIsNone(job.clean())

        def test_clean_invalid_callable(self):
            job = job_factory(self.JobModelClass)
            job.queue = list(settings.QUEUES)[0]
            job.callable = 'scheduler.tests.jobs.test_non_callable'
            with self.assertRaises(ValidationError):
                job.clean()

        def test_clean_invalid_queue(self):
            job = job_factory(self.JobModelClass)
            job.queue = 'xxxxxx'
            job.callable = 'scheduler.tests.jobs.test_job'
            with self.assertRaises(ValidationError):
                job.clean()

        def test_is_schedulable_already_scheduled(self):
            job = job_factory(self.JobModelClass, )
            job.schedule()
            self.assertTrue(job.is_scheduled())

        def test_is_schedulable_disabled(self):
            job = job_factory(self.JobModelClass)
            job.enabled = False
            self.assertFalse(job.enabled)

        def test_schedule(self):
            job = job_factory(self.JobModelClass, )
            self.assertTrue(job.is_scheduled())
            self.assertIsNotNone(job.job_id)

        def test_unschedulable(self):
            job = job_factory(self.JobModelClass, enabled=False)
            self.assertFalse(job.is_scheduled())
            self.assertIsNone(job.job_id)

        def test_unschedule(self):
            job = job_factory(self.JobModelClass, )
            self.assertTrue(job.unschedule())
            self.assertIsNone(job.job_id)

        def test_unschedule_not_scheduled(self):
            job = job_factory(self.JobModelClass, enabled=False)
            self.assertTrue(job.unschedule())
            self.assertIsNone(job.job_id)

        def test_save_enabled(self):
            job = job_factory(self.JobModelClass, )
            self.assertIsNotNone(job.job_id)

        def test_save_disabled(self):
            job = job_factory(self.JobModelClass, enabled=False)
            job.save()
            self.assertIsNone(job.job_id)

        def test_save_and_schedule(self):
            job = job_factory(self.JobModelClass, )
            self.assertIsNotNone(job.job_id)
            self.assertTrue(job.is_scheduled())

        def test_schedule2(self):
            job = job_factory(self.JobModelClass)
            job.queue = list(settings.QUEUES)[0]
            job.enabled = False
            job.scheduled_time = timezone.now() + timedelta(minutes=1)
            self.assertFalse(job.schedule())

        def test_delete_and_unschedule(self):
            job = job_factory(self.JobModelClass, )
            self.assertIsNotNone(job.job_id)
            self.assertTrue(job.is_scheduled())
            job.delete()
            self.assertFalse(job.is_scheduled())

        def test_job_create(self):
            prev_count = self.JobModelClass.objects.count()
            job_factory(self.JobModelClass)
            self.assertEqual(self.JobModelClass.objects.count(), prev_count + 1)

        def test_str(self):
            name = "test"
            job = job_factory(self.JobModelClass, name=name)
            self.assertEqual(f'{self.JobModelClass.__name__}[{name}={job.callable}()]', str(job))

        def test_callable_passthrough(self):
            job = job_factory(self.JobModelClass)
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.func, run_job)
            job_model, job_id = entry.args
            self.assertEqual(job_model, self.JobModelClass.__name__)
            self.assertEqual(job_id, job.id)

        def test_timeout_passthrough(self):
            job = job_factory(self.JobModelClass, timeout=500)
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.timeout, 500)

        def test_at_front_passthrough(self):
            job = job_factory(self.JobModelClass, at_front=True)
            queue = job.rqueue
            jobs_to_schedule = queue.scheduled_job_registry.get_job_ids()
            self.assertIn(job.job_id, jobs_to_schedule)

        def test_callable_result(self):
            job = job_factory(self.JobModelClass, )
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.perform(), 2)

        def test_callable_empty_args_and_kwargs(self):
            job = job_factory(self.JobModelClass, callable='scheduler.tests.jobs.test_args_kwargs')
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.perform(), 'test_args_kwargs()')

        def test_delete_args(self):
            job = job_factory(self.JobModelClass, )
            arg = jobarg_factory(JobArg, val='one', content_object=job)
            self.assertEqual(1, job.callable_args.count())
            arg.delete()
            self.assertEqual(0, job.callable_args.count())

        def test_delete_kwargs(self):
            job = job_factory(self.JobModelClass, )
            kwarg = jobarg_factory(JobKwarg, key='key1', arg_type='str', val='one', content_object=job)
            self.assertEqual(1, job.callable_kwargs.count())
            kwarg.delete()
            self.assertEqual(0, job.callable_kwargs.count())

        def test_parse_args(self):
            job = job_factory(self.JobModelClass, )
            date = timezone.now()
            jobarg_factory(JobArg, val='one', content_object=job)
            jobarg_factory(JobArg, arg_type='int', val=2, content_object=job)
            jobarg_factory(JobArg, arg_type='bool', val=True, content_object=job)
            jobarg_factory(JobArg, arg_type='bool', val=False, content_object=job)
            jobarg_factory(JobArg, arg_type='datetime', val=date, content_object=job)
            self.assertEqual(job.parse_args(), ['one', 2, True, False, date])

        def test_parse_kwargs(self):
            job = job_factory(self.JobModelClass, )
            date = timezone.now()
            jobarg_factory(JobKwarg, key='key1', arg_type='str', val='one', content_object=job)
            jobarg_factory(JobKwarg, key='key2', arg_type='int', val=2, content_object=job)
            jobarg_factory(JobKwarg, key='key3', arg_type='bool', val=True, content_object=job)
            jobarg_factory(JobKwarg, key='key4', arg_type='datetime', val=date, content_object=job)
            kwargs = job.parse_kwargs()
            self.assertEqual(kwargs, dict(key1='one', key2=2, key3=True, key4=date))

        def test_callable_args_and_kwargs(self):
            job = job_factory(self.JobModelClass, callable='scheduler.tests.jobs.test_args_kwargs')
            date = timezone.now()
            jobarg_factory(JobArg, arg_type='str', val='one', content_object=job)
            jobarg_factory(JobKwarg, key='key1', arg_type='int', val=2, content_object=job)
            jobarg_factory(JobKwarg, key='key2', arg_type='datetime', val=date, content_object=job)
            jobarg_factory(JobKwarg, key='key3', arg_type='bool', val=False, content_object=job)
            job.save()
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.perform(),
                             "test_args_kwargs('one', key1=2, key2={}, key3=False)".format(date))

        def test_function_string(self):
            job = job_factory(self.JobModelClass, )
            date = timezone.now()
            jobarg_factory(JobArg, arg_type='str', val='one', content_object=job)
            jobarg_factory(JobArg, arg_type='int', val='1', content_object=job)
            jobarg_factory(JobArg, arg_type='datetime', val=date, content_object=job)
            jobarg_factory(JobArg, arg_type='bool', val=True, content_object=job)
            jobarg_factory(JobKwarg, key='key1', arg_type='str', val='one', content_object=job)
            jobarg_factory(JobKwarg, key='key2', arg_type='int', val=2, content_object=job)
            jobarg_factory(JobKwarg, key='key3', arg_type='datetime', val=date, content_object=job)
            jobarg_factory(JobKwarg, key='key4', arg_type='bool', val=False, content_object=job)
            self.assertEqual(job.function_string(),
                             f"scheduler.tests.jobs.test_job('one', 1, {repr(date)}, True, "
                             f"key1='one', key2=2, key3={repr(date)}, key4=False)")

        def test_admin_list_view(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, )
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_list_view_delete_model(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, )
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.post(url, data={
                'action': 'delete_model',
                '_selected_action': [job.pk, ],
            })
            # assert
            self.assertEqual(302, res.status_code)

        def test_admin_run_job_now_enqueues_job_at(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, )
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.post(url, data={
                'action': 'enqueue_job_now',
                '_selected_action': [job.pk, ],
            })
            # assert
            self.assertEqual(302, res.status_code)
            job.refresh_from_db()
            queue = get_queue(job.queue)
            self.assertIn(job.job_id, queue.get_job_ids())

        def test_admin_change_view(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, )
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_change', args=[job.pk, ])
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_change_view__bad_redis_connection(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, queue='test2', instance_only=True)
            job.save(schedule_job=False)
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_change', args=[job.pk, ])
            # act
            res = self.client.get(url)
            # assert
            self.assertEqual(200, res.status_code)

        def test_admin_enqueue_job_now(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, )
            self.assertIsNotNone(job.job_id)
            self.assertTrue(job.is_scheduled())
            data = {
                'action': 'enqueue_job_now',
                '_selected_action': [job.id, ],
            }
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert part 1
            self.assertEqual(200, res.status_code)
            entry = _get_job_from_scheduled_registry(job)
            job_model, scheduled_job_id = entry.args
            self.assertEqual(job_model, job.JOB_TYPE)
            self.assertEqual(scheduled_job_id, job.id)
            self.assertEqual('scheduled', entry.get_status())
            self.assertTrue(has_execution_with_status(job, 'queued'))

            # act 2
            worker = create_worker('default', fork_job_execution=False, )
            worker.work(burst=True)

            # assert 2
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(job_model, job.JOB_TYPE)
            self.assertEqual(scheduled_job_id, job.id)
            self.assertTrue(has_execution_with_status(job, 'finished'))

        def test_admin_enable_job(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, enabled=False)
            self.assertIsNone(job.job_id)
            self.assertFalse(job.is_scheduled())
            data = {
                'action': 'enable_selected',
                '_selected_action': [job.id, ],
            }
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            job.refresh_from_db()
            self.assertTrue(job.enabled)
            self.assertTrue(job.is_scheduled())
            assert_response_has_msg(res, '1 job was successfully enabled and scheduled.')

        def test_admin_disable_job(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, enabled=True)
            job.save()
            data = {
                'action': 'disable_selected',
                '_selected_action': [job.id, ],
            }
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            self.assertTrue(job.is_scheduled())
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            job.refresh_from_db()
            self.assertFalse(job.is_scheduled())
            self.assertFalse(job.enabled)
            assert_response_has_msg(res, '1 job was successfully disabled and unscheduled.')

        def test_admin_single_delete(self):
            # arrange
            self.client.login(username='admin', password='admin')
            prev_count = self.JobModelClass.objects.count()
            job = job_factory(self.JobModelClass, )
            self.assertIsNotNone(job.job_id)
            self.assertTrue(job.is_scheduled())
            prev = len(_get_executions(job))
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_delete', args=[job.pk, ])
            data = {
                'post': 'yes',
            }
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            self.assertEqual(prev_count, self.JobModelClass.objects.count())
            self.assertEquals(prev - 1, len(_get_executions(job)))

        def test_admin_delete_selected(self):
            # arrange
            self.client.login(username='admin', password='admin')
            job = job_factory(self.JobModelClass, enabled=True)
            job.save()
            queue = get_queue(job.queue)
            scheduled_jobs = queue.scheduled_job_registry.get_job_ids()
            job_id = job.job_id
            self.assertIn(job_id, scheduled_jobs)
            data = {
                'action': 'delete_selected',
                '_selected_action': [job.id, ],
                'post': 'yes',
            }
            model = job._meta.model.__name__.lower()
            url = reverse(f'admin:scheduler_{model}_changelist')
            # act
            res = self.client.post(url, data=data, follow=True)
            # assert
            self.assertEqual(200, res.status_code)
            assert_response_has_msg(res, f'Successfully deleted 1 {self.JobModelClass._meta.verbose_name}.')
            self.assertIsNone(self.JobModelClass.objects.filter(id=job.id).first())
            scheduled_jobs = queue.scheduled_job_registry.get_job_ids()
            self.assertNotIn(job_id, scheduled_jobs)

    class TestSchedulableJob(TestBaseJob):
        # Currently ScheduledJob and RepeatableJob
        JobModelClass = ScheduledJob

        def test_schedule_time_utc(self):
            job = job_factory(self.JobModelClass)
            est = zoneinfo.ZoneInfo('US/Eastern')
            scheduled_time = datetime(2016, 12, 25, 8, 0, 0, tzinfo=est)
            job.scheduled_time = scheduled_time
            utc = zoneinfo.ZoneInfo('UTC')
            expected = scheduled_time.astimezone(utc).isoformat()
            self.assertEqual(expected, job._schedule_time().isoformat())

        def test_result_ttl_passthrough(self):
            job = job_factory(self.JobModelClass, result_ttl=500)
            entry = _get_job_from_scheduled_registry(job)
            self.assertEqual(entry.result_ttl, 500)


class TestScheduledJob(BaseTestCases.TestSchedulableJob):
    JobModelClass = ScheduledJob

    def test_clean(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        self.assertIsNone(job.clean())

    def test_unschedulable_old_job(self):
        job = job_factory(self.JobModelClass, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertFalse(job.is_scheduled())


class TestRepeatableJob(BaseTestCases.TestSchedulableJob):
    JobModelClass = RepeatableJob

    def test_unschedulable_old_job(self):
        job = job_factory(self.JobModelClass, scheduled_time=timezone.now() - timedelta(hours=1), repeat=0)
        self.assertFalse(job.is_scheduled())

    def test_schedulable_old_job_repeat_none(self):
        # If repeat is None, the job should be scheduled
        job = job_factory(self.JobModelClass, scheduled_time=timezone.now() - timedelta(hours=1), repeat=None)
        self.assertTrue(job.is_scheduled())

    def test_clean(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.result_ttl = -1
        self.assertIsNone(job.clean())

    def test_clean_seconds(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 60
        job.result_ttl = -1
        job.interval_unit = 'seconds'
        self.assertIsNone(job.clean())

    @override_settings(SCHEDULER_CONFIG={
        'SCHEDULER_INTERVAL': 10,
    })
    def test_clean_too_frequent(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 2  # Smaller than 10
        job.result_ttl = -1
        job.interval_unit = 'seconds'
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    def test_clean_not_multiple(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 121
        job.interval_unit = 'seconds'
        with self.assertRaises(ValidationError):
            job.clean_interval_unit()

    def test_clean_short_result_ttl(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.repeat = 1
        job.result_ttl = 3599
        job.interval_unit = 'hours'
        job.repeat = 42
        with self.assertRaises(ValidationError):
            job.clean_result_ttl()

    def test_clean_indefinite_result_ttl(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.result_ttl = -1
        job.interval_unit = 'hours'
        job.clean_result_ttl()

    def test_clean_undefined_result_ttl(self):
        job = job_factory(self.JobModelClass)
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        job.interval = 1
        job.interval_unit = 'hours'
        job.clean_result_ttl()

    def test_interval_seconds_weeks(self):
        job = job_factory(self.JobModelClass, interval=2, interval_unit='weeks')
        self.assertEqual(1209600.0, job.interval_seconds())

    def test_interval_seconds_days(self):
        job = job_factory(self.JobModelClass, interval=2, interval_unit='days')
        self.assertEqual(172800.0, job.interval_seconds())

    def test_interval_seconds_hours(self):
        job = job_factory(self.JobModelClass, interval=2, interval_unit='hours')
        self.assertEqual(7200.0, job.interval_seconds())

    def test_interval_seconds_minutes(self):
        job = job_factory(self.JobModelClass, interval=15, interval_unit='minutes')
        self.assertEqual(900.0, job.interval_seconds())

    def test_interval_seconds_seconds(self):
        job = RepeatableJob(interval=15, interval_unit='seconds')
        self.assertEqual(15.0, job.interval_seconds())

    def test_interval_display(self):
        job = job_factory(self.JobModelClass, interval=15, interval_unit='minutes')
        self.assertEqual(job.interval_display(), '15 minutes')

    def test_result_interval(self):
        job = job_factory(self.JobModelClass, )
        entry = _get_job_from_scheduled_registry(job)
        self.assertEqual(entry.meta['interval'], 3600)

    def test_repeat(self):
        job = job_factory(self.JobModelClass, repeat=10)
        entry = _get_job_from_scheduled_registry(job)
        self.assertEqual(entry.meta['repeat'], 10)

    def test_repeat_old_job_exhausted(self):
        base_time = timezone.now()
        job = job_factory(self.JobModelClass, scheduled_time=base_time - timedelta(hours=10), repeat=10)
        self.assertEqual(job.is_scheduled(), False)

    def test_repeat_old_job_last_iter(self):
        base_time = timezone.now()
        job = job_factory(self.JobModelClass, scheduled_time=base_time - timedelta(hours=9, minutes=30), repeat=10)
        self.assertEqual(job.repeat, 0)
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_old_job_remaining(self):
        base_time = timezone.now()
        job = job_factory(self.JobModelClass, scheduled_time=base_time - timedelta(minutes=30), repeat=5)
        self.assertEqual(job.repeat, 4)
        self.assertEqual(job.scheduled_time, base_time + timedelta(minutes=30))
        self.assertEqual(job.is_scheduled(), True)

    def test_repeat_none_interval_2_min(self):
        base_time = timezone.now()
        job = job_factory(self.JobModelClass, scheduled_time=base_time - timedelta(minutes=29), repeat=None)
        job.interval = 120
        job.interval_unit = 'seconds'
        job.schedule()
        self.assertTrue(job.scheduled_time > base_time)
        self.assertTrue(job.is_scheduled())

    def test_check_rescheduled_after_execution(self):
        job = job_factory(self.JobModelClass, scheduled_time=timezone.now() + timedelta(seconds=1))
        queue = job.rqueue
        first_run_id = job.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        job.refresh_from_db()
        self.assertTrue(job.is_scheduled())
        self.assertNotEquals(job.job_id, first_run_id)


class TestCronJob(BaseTestCases.TestBaseJob):
    JobModelClass = CronJob

    def test_clean(self):
        job = job_factory(self.JobModelClass)
        job.cron_string = '* * * * *'
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        self.assertIsNone(job.clean())

    def test_clean_cron_string_invalid(self):
        job = job_factory(self.JobModelClass)
        job.cron_string = 'not-a-cron-string'
        job.queue = list(settings.QUEUES)[0]
        job.callable = 'scheduler.tests.jobs.test_job'
        with self.assertRaises(ValidationError):
            job.clean_cron_string()

    def test_repeat(self):
        job = job_factory(self.JobModelClass, repeat=10)
        entry = _get_job_from_scheduled_registry(job)
        self.assertEqual(entry.meta['repeat'], 10)

    def test_check_rescheduled_after_execution(self):
        job = job_factory(self.JobModelClass, )
        queue = job.rqueue
        first_run_id = job.job_id
        entry = queue.fetch_job(first_run_id)
        queue.run_sync(entry)
        job.refresh_from_db()
        self.assertTrue(job.is_scheduled())
        self.assertNotEquals(job.job_id, first_run_id)



from typing import List, Any, Optional, Union

import django
from django.apps import apps
from redis import Redis
from redis.client import Pipeline
from rq import Worker
from rq.command import send_stop_job_command
from rq.decorators import job
from rq.exceptions import InvalidJobOperation
from rq.job import Job, JobStatus
from rq.job import get_current_job  # noqa
from rq.queue import Queue, logger
from rq.registry import (
    DeferredJobRegistry, FailedJobRegistry, FinishedJobRegistry,
    ScheduledJobRegistry, StartedJobRegistry, CanceledJobRegistry, BaseRegistry,
)
from rq.scheduler import RQScheduler
from rq.worker import WorkerStatus

from scheduler import settings

MODEL_NAMES = ['ScheduledJob', 'RepeatableJob', 'CronJob']

rq_job_decorator = job
ExecutionStatus = JobStatus
InvalidJobOperation = InvalidJobOperation


def as_text(v: Union[bytes, str]) -> Optional[str]:
    """Converts a bytes value to a string using `utf-8`.

    :param v: The value (bytes or string)
    :raises: ValueError: If the value is not bytes or string
    :returns: Either the decoded string or None
    """
    if v is None:
        return None
    elif isinstance(v, bytes):
        return v.decode('utf-8')
    elif isinstance(v, str):
        return v
    else:
        raise ValueError('Unknown type %r' % type(v))


def compact(lst: List[Any]) -> List[Any]:
    """Remove `None` values from an iterable object.
    :param lst: A list (or list-like) object
    :returns: The list without None values
    """
    return [item for item in lst if item is not None]


class JobExecution(Job):
    def __eq__(self, other):
        return isinstance(other, Job) and self.id == other.id

    @property
    def is_scheduled_job(self):
        return self.meta.get('scheduled_job_id', None) is not None

    def is_execution_of(self, scheduled_job):
        return (self.meta.get('job_type', None) == scheduled_job.JOB_TYPE
                and self.meta.get('scheduled_job_id', None) == scheduled_job.id)

    def stop_execution(self, connection: Redis):
        send_stop_job_command(connection, self.id)


class DjangoWorker(Worker):
    def __init__(self, *args, **kwargs):
        self.fork_job_execution = kwargs.pop('fork_job_execution', True)
        kwargs['job_class'] = JobExecution
        kwargs['queue_class'] = DjangoQueue
        super(DjangoWorker, self).__init__(*args, **kwargs)

    def __eq__(self, other):
        return (isinstance(other, Worker)
                and self.key == other.key
                and self.name == other.name)

    def __hash__(self):
        return hash((self.name, self.key, ','.join(self.queue_names())))

    def __str__(self):
        return f"{self.name}/{','.join(self.queue_names())}"

    def _start_scheduler(
            self,
            burst: bool = False,
            logging_level: str = "INFO",
            date_format: str = '%H:%M:%S',
            log_format: str = '%(asctime)s %(message)s',
    ) -> None:
        """Starts the scheduler process.
        This is specifically designed to be run by the worker when running the `work()` method.
        Instantiates the DjangoScheduler and tries to acquire a lock.
        If the lock is acquired, start scheduler.
        If worker is on burst mode just enqueues scheduled jobs and quits,
        otherwise, starts the scheduler in a separate process.


        :param burst (bool, optional): Whether to work on burst mode. Defaults to False.
        :param logging_level (str, optional): Logging level to use. Defaults to "INFO".
        :param date_format (str, optional): Date Format. Defaults to DEFAULT_LOGGING_DATE_FORMAT.
        :param log_format (str, optional): Log Format. Defaults to DEFAULT_LOGGING_FORMAT.
        """
        self.scheduler = DjangoScheduler(
            self.queues,
            connection=self.connection,
            logging_level=logging_level,
            date_format=date_format,
            log_format=log_format,
            serializer=self.serializer,
        )
        self.scheduler.acquire_locks()
        if self.scheduler.acquired_locks:
            if burst:
                self.scheduler.enqueue_scheduled_jobs()
                self.scheduler.release_locks()
            else:
                proc = self.scheduler.start()
                self._set_property('scheduler_pid', proc.pid)

    def execute_job(self, job: 'Job', queue: 'Queue'):
        if self.fork_job_execution:
            super(DjangoWorker, self).execute_job(job, queue)
        else:
            self.set_state(WorkerStatus.BUSY)
            self.perform_job(job, queue)
            self.set_state(WorkerStatus.IDLE)

    def work(self, **kwargs) -> bool:
        kwargs.setdefault('with_scheduler', True)
        return super(DjangoWorker, self).work(**kwargs)

    def _set_property(self, prop_name: str, val, pipeline: Optional[Pipeline] = None):
        connection = pipeline if pipeline is not None else self.connection
        if val is None:
            connection.hdel(self.key, prop_name)
        else:
            connection.hset(self.key, prop_name, val)

    def _get_property(self, prop_name: str, pipeline: Optional[Pipeline] = None):
        connection = pipeline if pipeline is not None else self.connection
        return as_text(connection.hget(self.key, prop_name))

    def scheduler_pid(self) -> int:
        pid = self.connection.get(RQScheduler.get_locking_key(self.queues[0].name))
        return int(pid.decode()) if pid is not None else None


class DjangoQueue(Queue):
    REGISTRIES = dict(
        finished='finished_job_registry',
        failed='failed_job_registry',
        scheduled='scheduled_job_registry',
        started='started_job_registry',
        deferred='deferred_job_registry',
        canceled='canceled_job_registry',
    )
    """
    A subclass of RQ's QUEUE that allows jobs to be stored temporarily to be
    enqueued later at the end of Django's request/response cycle.
    """

    def __init__(self, *args, **kwargs):
        kwargs['job_class'] = JobExecution
        super(DjangoQueue, self).__init__(*args, **kwargs)

    def get_registry(self, name: str) -> Union[None, BaseRegistry, 'DjangoQueue']:
        name = name.lower()
        if name == 'queued':
            return self
        elif name in DjangoQueue.REGISTRIES:
            return getattr(self, DjangoQueue.REGISTRIES[name])
        return None

    @property
    def finished_job_registry(self):
        return FinishedJobRegistry(self.name, self.connection)

    @property
    def started_job_registry(self):
        return StartedJobRegistry(self.name, self.connection, job_class=JobExecution, )

    @property
    def deferred_job_registry(self):
        return DeferredJobRegistry(self.name, self.connection, job_class=JobExecution, )

    @property
    def failed_job_registry(self):
        return FailedJobRegistry(self.name, self.connection, job_class=JobExecution, )

    @property
    def scheduled_job_registry(self):
        return ScheduledJobRegistry(self.name, self.connection, job_class=JobExecution, )

    @property
    def canceled_job_registry(self):
        return CanceledJobRegistry(self.name, self.connection, job_class=JobExecution, )

    def get_all_job_ids(self) -> List[str]:
        res = list()
        res.extend(self.get_job_ids())
        res.extend(self.finished_job_registry.get_job_ids())
        res.extend(self.started_job_registry.get_job_ids())
        res.extend(self.deferred_job_registry.get_job_ids())
        res.extend(self.failed_job_registry.get_job_ids())
        res.extend(self.scheduled_job_registry.get_job_ids())
        res.extend(self.canceled_job_registry.get_job_ids())
        return res

    def get_all_jobs(self):
        job_ids = self.get_all_job_ids()
        return compact([self.fetch_job(job_id) for job_id in job_ids])

    def clean_registries(self):
        self.started_job_registry.cleanup()
        self.failed_job_registry.cleanup()
        self.finished_job_registry.cleanup()

    def remove_job_id(self, job_id: str):
        self.connection.lrem(self.key, 0, job_id)

    def last_job_id(self):
        return self.connection.lindex(self.key, 0)


class DjangoScheduler(RQScheduler):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault('interval', settings.SCHEDULER_CONFIG['SCHEDULER_INTERVAL'])
        super(DjangoScheduler, self).__init__(*args, **kwargs)

    @staticmethod
    def reschedule_all_jobs():
        for model_name in MODEL_NAMES:
            model = apps.get_model(app_label='scheduler', model_name=model_name)
            enabled_jobs = model.objects.filter(enabled=True)
            unscheduled_jobs = filter(lambda j: j.ready_for_schedule(), enabled_jobs)
            for item in unscheduled_jobs:
                logger.debug(f"Rescheduling {str(item)}")
                item.save()

    def work(self):
        django.setup()
        super(DjangoScheduler, self).work()

    def enqueue_scheduled_jobs(self):
        self.reschedule_all_jobs()
        super(DjangoScheduler, self).enqueue_scheduled_jobs()



import math
import uuid
from datetime import timedelta
from typing import Dict

import croniter
from django.apps import apps
from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import ValidationError
from django.db import models
from django.templatetags.tz import utc
from django.urls import reverse
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from scheduler import settings
from scheduler import tools
from scheduler.models.args import JobArg, JobKwarg
from scheduler.queues import get_queue, logger
from scheduler.rq_classes import DjangoQueue

SCHEDULER_INTERVAL = settings.SCHEDULER_CONFIG['SCHEDULER_INTERVAL']


def callback_save_job(job, connection, result, *args, **kwargs):
    model_name = job.meta.get('job_type', None)
    if model_name is None:
        return
    model = apps.get_model(app_label='scheduler', model_name=model_name)
    scheduled_job = model.objects.filter(job_id=job.id).first()
    if scheduled_job is not None:
        scheduled_job.unschedule()
        scheduled_job.schedule()


class BaseJob(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    QUEUES = [(key, key) for key in settings.QUEUES.keys()]
    JOB_TYPE = 'BaseJob'
    name = models.CharField(
        _('name'), max_length=128, unique=True,
        help_text='Name of the job.', )
    callable = models.CharField(_('callable'), max_length=2048)
    callable_args = GenericRelation(JobArg, related_query_name='args')
    callable_kwargs = GenericRelation(JobKwarg, related_query_name='kwargs')
    enabled = models.BooleanField(
        _('enabled'), default=True,
        help_text=_('Should job be scheduled? This field is useful to keep '
                    'past jobs that should no longer be scheduled'),
    )
    queue = models.CharField(
        _('queue'), max_length=255, choices=QUEUES,
        help_text=_('Queue name'), )
    job_id = models.CharField(
        _('job id'), max_length=128, editable=False, blank=True, null=True,
        help_text=_('Current job_id on queue'))
    repeat = models.PositiveIntegerField(
        _('repeat'), blank=True, null=True,
        help_text=_('Number of times to run the job. Leaving this blank means it will run forever.'), )
    at_front = models.BooleanField(
        _('At front'), default=False, blank=True, null=True,
        help_text=_('When queuing the job, add it in the front of the queue'), )
    timeout = models.IntegerField(
        _('timeout'), blank=True, null=True,
        help_text=_("Timeout specifies the maximum runtime, in seconds, for the job "
                    "before it'll be considered 'lost'. Blank uses the default "
                    "timeout."), )
    result_ttl = models.IntegerField(
        _('result ttl'), blank=True, null=True,
        help_text=mark_safe(
            """The TTL value (in seconds) of the job result.<br/>
               -1: Result never expires, you should delete jobs manually. <br/>
                0: Result gets deleted immediately. <br/>
                >0: Result expires after n seconds."""), )

    def callable_func(self):
        """Translate callable string to callable"""
        return tools.callable_func(self.callable)

    @admin.display(boolean=True, description=_('is scheduled?'))
    def is_scheduled(self) -> bool:
        """Check whether job is queued/scheduled to be executed"""
        if not self.job_id:  # no job_id => is not scheduled
            return False
        # check whether job_id is in scheduled/enqueued/active jobs
        scheduled_jobs = self.rqueue.scheduled_job_registry.get_job_ids()
        enqueued_jobs = self.rqueue.get_job_ids()
        active_jobs = self.rqueue.started_job_registry.get_job_ids()
        res = ((self.job_id in scheduled_jobs)
               or (self.job_id in enqueued_jobs)
               or (self.job_id in active_jobs))
        # If the job_id is not scheduled/enqueued/started,
        # update the job_id to None. (The job_id belongs to a previous run which is completed)
        if not res:
            self.job_id = None
            super(BaseJob, self).save()
        return res

    @admin.display(description='Callable')
    def function_string(self) -> str:
        args = self.parse_args()
        args_list = [repr(arg) for arg in args]
        kwargs = self.parse_kwargs()
        kwargs_list = [k + '=' + repr(v) for (k, v) in kwargs.items()]
        return self.callable + f"({', '.join(args_list + kwargs_list)})"

    def parse_args(self):
        """Parse args for running job"""
        args = self.callable_args.all()
        return [arg.value() for arg in args]

    def parse_kwargs(self):
        """Parse kwargs for running job"""
        kwargs = self.callable_kwargs.all()
        return dict([kwarg.value() for kwarg in kwargs])

    def _next_job_id(self):
        addition = uuid.uuid4().hex[-10:]
        name = self.name.replace('/', '.')
        return f'{self.queue}:{name}:{addition}'

    def _enqueue_args(self) -> Dict:
        """args for DjangoQueue.enqueue.
        Set all arguments for DjangoQueue.enqueue/enqueue_at.
        Particularly:
        - set job timeout and ttl
        - ensure a callback to reschedule job next iteration.
        - set job-id to proper format
        - set job meta
        """
        res = dict(
            meta=dict(
                repeat=self.repeat,
                job_type=self.JOB_TYPE,
                scheduled_job_id=self.id,
            ),
            on_success=callback_save_job,
            on_failure=callback_save_job,
            job_id=self._next_job_id(),
        )
        if self.at_front:
            res['at_front'] = self.at_front
        if self.timeout:
            res['job_timeout'] = self.timeout
        if self.result_ttl is not None:
            res['result_ttl'] = self.result_ttl
        return res

    @property
    def rqueue(self) -> DjangoQueue:
        """Returns django-queue for job
        """
        return get_queue(self.queue)

    def ready_for_schedule(self) -> bool:
        """Is job ready to be scheduled?

        If job is already scheduled or disabled, then it is not
        ready to be scheduled.

        :returns: True if job is ready to be scheduled.
        """
        if self.is_scheduled():
            logger.debug(f'Job {self.name} already scheduled')
            return False
        if not self.enabled:
            logger.debug(f'Job {str(self)} disabled, enable job before scheduling')
            return False
        return True

    def schedule(self) -> bool:
        """Schedule job to run.
        :returns: True if job was scheduled, False otherwise.
        """
        if not self.ready_for_schedule():
            return False
        schedule_time = self._schedule_time()
        kwargs = self._enqueue_args()
        job = self.rqueue.enqueue_at(
            schedule_time,
            tools.run_job,
            args=(self.JOB_TYPE, self.id),
            **kwargs, )
        self.job_id = job.id
        super(BaseJob, self).save()
        return True

    def enqueue_to_run(self) -> bool:
        """Enqueue job to run now.
        """
        kwargs = self._enqueue_args()
        job = self.rqueue.enqueue(
            tools.run_job,
            args=(self.JOB_TYPE, self.id),
            **kwargs,
        )
        self.job_id = job.id
        super(BaseJob, self).save()
        return True

    def unschedule(self) -> bool:
        """Remove job from django-queue.

        If job is queued to be executed or scheduled to be executed, it will remove it.
        """
        queue = self.rqueue
        if self.is_scheduled():
            queue.remove(self.job_id)
            queue.scheduled_job_registry.remove(self.job_id)
        self.job_id = None
        super(BaseJob, self).save()
        return True

    def _schedule_time(self):
        return utc(self.scheduled_time)

    def to_dict(self) -> Dict:
        """Export model to dictionary, so it can be saved as external file backup
        """
        res = dict(
            model=self.JOB_TYPE,
            name=self.name,
            callable=self.callable,
            callable_args=[
                dict(arg_type=arg.arg_type, val=arg.val, )
                for arg in self.callable_args.all()],
            callable_kwargs=[
                dict(arg_type=arg.arg_type, key=arg.key, val=arg.val, )
                for arg in self.callable_kwargs.all()],
            enabled=self.enabled,
            queue=self.queue,
            repeat=self.repeat,
            at_front=self.at_front,
            timeout=self.timeout,
            result_ttl=self.result_ttl,
            cron_string=getattr(self, 'cron_string', None),
            scheduled_time=self._schedule_time().isoformat(),
            interval=getattr(self, 'interval', None),
            interval_unit=getattr(self, 'interval_unit', None),
        )
        return res

    def get_absolute_url(self):
        model = self._meta.model.__name__.lower()
        return reverse(f'admin:scheduler_{model}_change', args=[self.id, ])

    def __str__(self):
        func = self.function_string()
        return f'{self.JOB_TYPE}[{self.name}={func}]'

    def save(self, **kwargs):
        schedule_job = kwargs.pop('schedule_job', True)
        update_fields = kwargs.get('update_fields', None)
        if update_fields:
            kwargs['update_fields'] = set(update_fields).union({'modified'})
        super(BaseJob, self).save(**kwargs)
        if schedule_job:
            self.schedule()
            super(BaseJob, self).save()

    def delete(self, **kwargs):
        self.unschedule()
        super(BaseJob, self).delete(**kwargs)

    def clean_callable(self):
        try:
            tools.callable_func(self.callable)
        except Exception:
            raise ValidationError({
                'callable': ValidationError(
                    _('Invalid callable, must be importable'), code='invalid')
            })

    def clean_queue(self):
        queue_keys = settings.QUEUES.keys()
        if self.queue not in queue_keys:
            raise ValidationError({
                'queue': ValidationError(
                    _('Invalid queue, must be one of: {}'.format(
                        ', '.join(queue_keys))), code='invalid')
            })

    def clean(self):
        self.clean_queue()
        self.clean_callable()

    class Meta:
        abstract = True


class ScheduledTimeMixin(models.Model):
    scheduled_time = models.DateTimeField(_('scheduled time'))

    class Meta:
        abstract = True


class ScheduledJob(ScheduledTimeMixin, BaseJob):
    repeat = None
    JOB_TYPE = 'ScheduledJob'

    def ready_for_schedule(self) -> bool:
        return (super(ScheduledJob, self).ready_for_schedule()
                and (self.scheduled_time is None
                     or self.scheduled_time >= timezone.now()))

    class Meta:
        verbose_name = _('Scheduled Job')
        verbose_name_plural = _('Scheduled Jobs')
        ordering = ('name',)


class RepeatableJob(ScheduledTimeMixin, BaseJob):
    class TimeUnits(models.TextChoices):
        SECONDS = 'seconds', _('seconds')
        MINUTES = 'minutes', _('minutes')
        HOURS = 'hours', _('hours')
        DAYS = 'days', _('days')
        WEEKS = 'weeks', _('weeks')

    interval = models.PositiveIntegerField(_('interval'))
    interval_unit = models.CharField(
        _('interval unit'), max_length=12, choices=TimeUnits.choices, default=TimeUnits.HOURS
    )
    JOB_TYPE = 'RepeatableJob'

    def clean(self):
        super(RepeatableJob, self).clean()
        self.clean_interval_unit()
        self.clean_result_ttl()

    def clean_interval_unit(self):
        if SCHEDULER_INTERVAL > self.interval_seconds():
            raise ValidationError(
                _("Job interval is set lower than %(queue)r queue's interval. "
                  "minimum interval is %(interval)"),
                code='invalid',
                params={'queue': self.queue, 'interval': SCHEDULER_INTERVAL})
        if self.interval_seconds() % SCHEDULER_INTERVAL:
            raise ValidationError(
                _("Job interval is not a multiple of rq_scheduler's interval frequency: %(interval)ss"),
                code='invalid',
                params={'interval': SCHEDULER_INTERVAL})

    def clean_result_ttl(self) -> None:
        """
        Throws an error if there are repeats left to run and the result_ttl won't last until the next scheduled time.
        :return: None
        """
        if self.result_ttl and self.result_ttl != -1 and self.result_ttl < self.interval_seconds() and self.repeat:
            raise ValidationError(
                _("Job result_ttl must be either indefinite (-1) or "
                  "longer than the interval, %(interval)s seconds, to ensure rescheduling."),
                code='invalid',
                params={'interval': self.interval_seconds()}, )

    def interval_display(self):
        return '{} {}'.format(self.interval, self.get_interval_unit_display())

    def interval_seconds(self):
        kwargs = {self.interval_unit: self.interval, }
        return timedelta(**kwargs).total_seconds()

    def _enqueue_args(self):
        res = super(RepeatableJob, self)._enqueue_args()
        res['meta']['interval'] = self.interval_seconds()
        return res

    def ready_for_schedule(self):
        if super(RepeatableJob, self).ready_for_schedule() is False:
            return False
        if self.scheduled_time < timezone.now():
            gap = math.ceil((timezone.now().timestamp() - self.scheduled_time.timestamp()) / self.interval_seconds())
            if self.repeat is None or self.repeat >= gap:
                self.scheduled_time += timedelta(seconds=self.interval_seconds() * gap)
                self.repeat = (self.repeat - gap) if self.repeat is not None else None

        if self.scheduled_time < timezone.now():
            return False
        return True

    class Meta:
        verbose_name = _('Repeatable Job')
        verbose_name_plural = _('Repeatable Jobs')
        ordering = ('name',)


class CronJob(BaseJob):
    JOB_TYPE = 'CronJob'

    cron_string = models.CharField(
        _('cron string'), max_length=64,
        help_text=mark_safe(
            '''Define the schedule in a crontab like syntax.
            Times are in UTC. Use <a href="https://crontab.guru/">crontab.guru</a> to create a cron string.''')
    )

    def clean(self):
        super(CronJob, self).clean()
        self.clean_cron_string()

    def clean_cron_string(self):
        try:
            croniter.croniter(self.cron_string)
        except ValueError as e:
            raise ValidationError({'cron_string': ValidationError(_(str(e)), code='invalid')})

    def _schedule_time(self):
        return tools.get_next_cron_time(self.cron_string)

    class Meta:
        verbose_name = _('Cron Job')
        verbose_name_plural = _('Cron Jobs')
        ordering = ('name',)



import importlib
import os

import croniter
from django.apps import apps
from django.templatetags.tz import utc
from django.utils import timezone

from scheduler.queues import get_queues, logger, get_queue
from scheduler.rq_classes import DjangoWorker, MODEL_NAMES
from scheduler.settings import get_config


def callable_func(callable_str: str):
    path = callable_str.split('.')
    module = importlib.import_module('.'.join(path[:-1]))
    func = getattr(module, path[-1])
    if callable(func) is False:
        raise TypeError("'{}' is not callable".format(callable_str))
    return func


def get_next_cron_time(cron_string) -> timezone.datetime:
    """Calculate the next scheduled time by creating a crontab object
    with a cron string"""
    now = timezone.now()
    itr = croniter.croniter(cron_string, now)
    return utc(itr.get_next(timezone.datetime))


def get_scheduled_job(task_model: str, task_id: int):
    if task_model not in MODEL_NAMES:
        raise ValueError(f'Job Model {task_model} does not exist, choices are {MODEL_NAMES}')
    model = apps.get_model(app_label='scheduler', model_name=task_model)
    task = model.objects.filter(id=task_id).first()
    if task is None:
        raise ValueError(f'Job {task_model}:{task_id} does not exit')
    return task


def run_job(task_model: str, task_id: int):
    """Run a scheduled job
    """
    scheduled_job = get_scheduled_job(task_model, task_id)
    logger.debug(f'Running task {str(scheduled_job)}')
    args = scheduled_job.parse_args()
    kwargs = scheduled_job.parse_kwargs()
    res = scheduled_job.callable_func()(*args, **kwargs)
    return res


def _calc_worker_name(existing_worker_names):
    hostname = os.uname()[1]
    c = 1
    worker_name = f'{hostname}-worker.{c}'
    while worker_name in existing_worker_names:
        c += 1
        worker_name = f'{hostname}-worker.{c}'
    return worker_name


def create_worker(*queue_names, **kwargs):
    """
    Returns a Django worker for all queues or specified ones.
    """

    queues = get_queues(*queue_names)
    existing_workers = DjangoWorker.all(connection=queues[0].connection)
    existing_worker_names = set(map(lambda w: w.name, existing_workers))
    kwargs['fork_job_execution'] = not get_config('FAKEREDIS')
    if kwargs.get('name', None) is None:
        kwargs['name'] = _calc_worker_name(existing_worker_names)

    kwargs['name'] = kwargs['name'].replace('/', '.')
    worker = DjangoWorker(queues, connection=queues[0].connection, **kwargs)
    return worker


def get_job_executions(queue_name, scheduled_job):
    queue = get_queue(queue_name)
    job_list = queue.get_all_jobs()
    res = list(filter(lambda j: j.is_execution_of(scheduled_job), job_list))
    return res
